/*
   =============================================
    filtbank.c: C code for auditory filter bank
   =============================================
  Filtbank uses file input/output, writing full filter bank output to file so
    that f_align can rejig it.

  USE :  filtbank  -i speech file name -o filter bank output file name 
  filtbank: is compiled with (makefile for WATCOM 'C'):

  rwfiles.c - filehandling 
  initial4.c - initialises arrays and structures
  filtunix.c - contains recursive filter code
  formulae.c - Erb and Erb Scale formulae, new (1990) and old (1986)
  rwbiquad.c - read/write biquad.tst file
  cmplxmth.c - complex math operations

Originally M.A.Stone August 1990.
Mods for recruitment simulation B.R.Glasberg 1993-1995

*/

#include "recwide4.h"
#include "proto.h"
#define SIGMAX  32767.
#define SIGMIN -32767. /* helps with rectifying if symmetric */

/* filter bank on number time samples:
 op must be N_OF_CHANS times bigger than ip
  - this is OUTBUFSIZE in recwide4.h */

/* static data areas defined here */
struct Biquad filter_coeffs[N_OF_CHANS];  /* filterbank coefficients structs */
struct Hpbessel hpcoeffts[No_to_HPass]; /* hp bessel coefficients */
Filter_State  bank_state[N_OF_CHANS][XiYj];  /* state vectors of each filter */
float in[BUFFSIZE], out[OUTBUFSIZE], ch_gain = 1.0;  /* filterbank io */
short folder[BUFFSIZE], bankvect[OUTBUFSIZE];  /*file read/write buffer*/

/*---------------------------------------------------*/
void main(int argc,char **argv) /* MAIN starts here. */
{
/* internal variables */

  long    i, grabsize, outitems, options, scale, blockno;
  short   *shrt_ptr;
  float   *flt_ptr, temp;
  char    *inputname, *outputname;
  FILE    *reading_from, *writing_to;
  long    overflow = 0L, undrflow = 0L;

/* Parse Command Line */
  scale = 0; options = 2; /* 2 compulsory options */

  for (i=1; i < argc; i++) {
      switch (*argv[i]) {
    case '-':
	switch (*(argv[i] + 1)) {
      case 'i': inputname  = argv[++i]; options--; break;
      case 'o': outputname = argv[++i]; options--; break;
      case 'g': scale = 1; ch_gain = atof(argv[++i]); break;
      /*break;*/
	}
      }
  }
 if (options) 
    error("use: widefilt -i Ip file -o Op file [-g channel gain]\n\n",NULLPTR);
  if( scale)   printf(": Output scaled by %6.2f\n",ch_gain);

/* Open all the Files Needed */
  if ( (reading_from = fopen (inputname, "rb")) == NULLPTR)
	error (" widefilt: Can't open INPUT file: %s\n", inputname);

  if ( (writing_to = fopen (outputname, "wb")) == NULLPTR)
	error (" widefilt: Can't Create OUTPUT file: %s\n", outputname);

/*--------------------------------------*/
/* Initialise filters and state vectors */

  zero_array(in, BUFFSIZE); zero_array(out, (OUTBUFSIZE));
  intlse_filter_states(bank_state[0]);

  calc_hp_coeffts(hpcoeffts); /* 3 highest channels have high-pass Bessel in series */
/* do this 1st then add delays in calc_filter_data */

  calculate_filter_data(filter_coeffs,YES); /* set to save filterbank data */

/*  calculate_filter_data(filter_coeffs,NO); *//* set NOT to save filterbank data */

/*------------------------------------------------------------------------*/
/* enter main part of loop; repeat in BUFFSIZE chunks until end of file   */

grabsize = BUFFSIZE; blockno = 0;

printf("\nfilter bank: ");
while( grabsize == BUFFSIZE ){
  printf(".");fflush(stdout);
  grabsize = readnextsect(reading_from,folder);
  convert_to_float(folder,in,grabsize);

    aud_filt_bank(in,out,grabsize); /* auditory filtering */
    outitems = N_OF_CHANS * grabsize;

    if(scale){   /* optional scale output of each channel by input variable */
      flt_ptr = out; shrt_ptr = bankvect;
      for(i = 0; i++ < outitems; ) {
	(temp = (*(flt_ptr++) * ch_gain));
	if     (temp > SIGMAX){ overflow++; *(shrt_ptr++) = (short)SIGMAX;}
	else if(temp < SIGMIN){ undrflow++; *(shrt_ptr++) = (short)SIGMIN;}
	else *(shrt_ptr++) = (short) temp;
      }
    }
    else convert_to_short(out, bankvect, outitems);
    if (fwrite (bankvect, sizeof(short), outitems, writing_to) == 0)
	error("widefilt: Error in WRITING",NULLPTR);
  }
  fclose(reading_from); fclose(writing_to);
  printf("\nfilter bank:  normal termination\n"); 
/*  printf("Short integer output file: %s\n",outputname);*/
  if((overflow > 0) || (undrflow > 0))
    printf(" %7ld overflows, %7ld underflows in output file %s\n", overflow, undrflow, outputname);
}
/* ----------  end  of filtbank.c  ------------------------  */
