/*
	 ==============================================================
	 fir_corr.c: C code to apply compensatory fir filter to audio.
	 Origin is time-shifted by half-aperture width...............
	 MAStone, 26-May-2001: for use in BRG recruitment simulation
	 ===============================================================
*/
#define WATCOMC /* which compiler to use */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>

#ifdef WATCOMC
	#include <malloc.h>
	#define SGMNT 16384  /* how many samples to process */
#else
	#include <alloc.h>
	#define SGMNT 8192  /* how many samples to process */
#endif


#define MAX_HALF 256	/* maximum halfwidth of fir filter */
#define SHORTS sizeof(short)
#define FLOATS sizeof(float)
#define SIGMAX  32767.
#define SIGMIN -32767. /* helps with rectifying if symmetric */
#define NULLPTR (void *)0
#define UPDATE 3 /* (how many buffers-1) before update sample numbers to screen */ 


/*--------------------------------------------------------------*/
void error(char *msg1, char *msg2){/* print user error and quit */
	printf("\nERROR: Quitting FIR_CORR due to:\n\t%s %s\n",msg1,msg2);
	exit(1);
}
/*--------------------------------------------------------------*/
/* static data areas defined here */
float filter[2*MAX_HALF+1]; /* static coeficient space */
short in[SGMNT], out[SGMNT]; /* filter input/output */

/*------------------------------------------------*/
void zero_short_array(short *elemnt, short howmany){
/*----- Initialise an array of short to zero */
register short i;
for(i = howmany; i-- > 0; ) *(elemnt++) = 0;
}
/*-------------------------------------------*/
void zero_float_array(float *elemnt, short howmany){
/*----- Initialise an array of float to zero */
register short i;
for(i = howmany; i-- > 0; ) *(elemnt++) = 0.0;
}
/*---------------------------------------------------------------*/
void convert_to_float(short *input, float *output, short howmany){
/*--------------------------------- convert input array to float */
register short i;
for(i = 0; i++ < howmany; ) *(output++) = (float) *(input++);
}
/*---------------------------------------------------------------*/
void convert_to_short(float *input, short *output, short howmany){
/*--------------------------------- convert input array to short */
register short i;
for(i = 0; i++ < howmany; ) *(output++) = (short) *(input++);
}
/*---------------------------------------------------------------------------------*/
short rd_fir_file(FILE *fir_file, float *coeffs, float scale){ /* read filter file */
short idx, n_cfs;
char sntnc[80];
float tmp;
fscanf(fir_file, "%[^\n]", &sntnc); /* first line has number of coefficients */
fscanf(fir_file, "%d", &n_cfs); /* second line has number of coefficients */
printf("\n\tFIR filter [%s], %d coeffs\n",sntnc,n_cfs);
if (n_cfs > (2*MAX_HALF+1)){
	sprintf(sntnc,"%d but can only handle %d coefficients",n_cfs,(2*MAX_HALF+1));
	error("filter file too big:",sntnc);
}
for (idx = 0; idx < n_cfs; idx ++){
	if(fscanf(fir_file, "%f", &tmp) == EOF) break;
	*(coeffs+idx) = tmp*scale;
}
if (idx != n_cfs) {
	sprintf(sntnc,"%d but only found %d coefficients",n_cfs,idx);
	error("filter file incorrect size:",sntnc);
}
printf("\nAll filtering will be scaled by factor of %6.3f\n",scale);
return(n_cfs);
}
/*-----------------------------------------------------------------------------*/
float do_fir(float *in, float *fir_filt,short ntaps){ /* perform fir filter fn */
float tmp;
short idx;
	tmp = 0.;
	for (idx = 0; idx < ntaps; idx++) tmp += (*(in++) * *(fir_filt++));
return(tmp);
}
/*-------------------------------------------------------*/
void main(short argc, char **argv){ /* MAIN starts here. */
/* internal variables */
	short   grabsize, ix, options, *shrt_ptr, fir_size;
	float   *flt_ptr_i, *flt_ptr_o, temp;
	float   *filter_ip, gain ; /* float version of buffer of each filter */
	char    *inputname, *outputname, *fir_name;
	FILE    *reading_from, *writing_to, *filter_from;
	long    overflow = 0L, undrflow = 0L, sofar = 0L, update = 0L;

#ifdef WATCOMC
//	printf("\n\t\tWATCOM C Compiler\n");
#else
//	printf("\n\t\t TURBO C Compiler\n");
#endif

/* Parse Command Line */
	options = 3; gain = 1.0;
	for (ix=1; ix < argc; ix++) {
			switch (*argv[ix]) {
		case '-':
				switch (*(argv[ix] + 1)) {
			case 'i': inputname  = argv[++ix]; options--; break;
			case 'o': outputname = argv[++ix]; options--; break;
			case 'f': fir_name   = argv[++ix]; options--; break;
			case 'g': gain = atof( argv[++ix]); break;

				}
			}
	}

	if (options){
		error("Insufficient options\n", "fir_corr -i Ip file -o Op file -f filter coeff file[-g gain scaling]\n");
	}
	else {
		printf("\nFIR_CORR: applies max %d-tap FIR filter to 16 bit files", 2*MAX_HALF+1);
	}

/* Open all the Files Needed */
	if ( (reading_from = fopen (inputname, "rb")) == NULLPTR)
			error ("Can't open INPUT file: ", inputname);
	else if ( (writing_to = fopen (outputname, "wb")) == NULLPTR)
			error ("Can't Create OUTPUT file:", outputname);
	else if ( (filter_from = fopen (fir_name, "rt")) == NULLPTR)
			error ("Can't open FILTER file:", fir_name);

	if((filter_ip = malloc((SGMNT+2*MAX_HALF)*FLOATS)) == NULLPTR)
		error("out of MEMORY for float INPUT area", NULLPTR);

/*--------------------------------------*/
/* Initialise areas */
	zero_short_array(in, SGMNT); zero_short_array(out, SGMNT);
	zero_float_array(filter_ip, (SGMNT+2*MAX_HALF));
	fir_size = rd_fir_file(filter_from, filter, gain);  /* read filter file and scale */
	fclose(filter_from); /* and close it */
/*------------------------------------------------------------------------*/
/* enter main part of loop; repeat in SGMNT chunks until end of file   */

grabsize = SGMNT;

while( grabsize == SGMNT ){
	shrt_ptr = out; flt_ptr_i = filter_ip; /* initialise */
	if (( grabsize = fread (in, SHORTS, SGMNT, reading_from)) < 0)
		error(" Error in READING data.",NULLPTR);
	convert_to_float(in, filter_ip+fir_size-1, grabsize);

	for (ix = 0; ix < grabsize; ix ++){
		temp = do_fir(flt_ptr_i++, filter, fir_size); /* FIR filtering */
		if (temp > SIGMAX){ overflow++; temp = SIGMAX;}
		else if(temp < SIGMIN){ undrflow++; temp = SIGMIN;}
		*(shrt_ptr++) = (short) temp;
	}
	if (fwrite (out, SHORTS, grabsize, writing_to) <= 0)
		error(" Error in WRITING output data",NULLPTR);
	/* shuffle data down through array ready for next loop */
	flt_ptr_i = filter_ip+grabsize;
	flt_ptr_o = filter_ip;
	for (ix = 0; ix < fir_size-1; ix++){
		*(flt_ptr_o++) = *flt_ptr_i;
		*(flt_ptr_i++) = 0.; /* safety to obliterate old data in wrong place */
	}
	sofar += grabsize; update += 1;
	if ((update & UPDATE) == UPDATE) printf("\r%7ld",sofar); /* don't print so often to the screen */
}
	printf("\r%7ld",sofar);
	fclose(reading_from); fclose(writing_to); free(filter_ip);
	printf("\nFIR_CORR: normal termination to produce:\n");
	printf(" 16-bit Op file: %7ld overflows, %7ld underflows\n", overflow, undrflow);
}
