/*
  f_align.c subroutine to time align the filter bank output in a file.
    M.A.Stone November 1990

  Assumes that there will be N_OF_CHANS filters in use: looks to biquadratic
  filter file to extract delays for each filter. Time alignment occurs by use
  of two blocks of data, the processed block being one block behind the last
  block read. This means that the maximum look ahead is defined by INSEG/2, at
  present this is 256 samples at 16kHz, or 16ms. Unlikely to be able to use
  more unless compiler can handle malloc() for blocks > 64K.
  LINK with RWBIQUAD - reads biquad.tst file
*/
#include <malloc.h>
#include <string.h>
#include "recwide4.h"
#define INSEG 512
#define OUTSEG INSEG/2

#include "proto.h"

struct Biquad filter_coeffs[N_OF_CHANS], *filt_i;
short  *strt0, *strt1; /* frequently used data, keep global */

/*-----------------------------------------------------*/
void err_rpt(char *msg)  /* print diagnostic and abort */
{  printf("\n** error in f_align:\n"); printf(msg); exit(1); }
/*-------------*/
void sendhelp(void){
err_rpt("use:  f_align -i Ip file -o Op file\n Time aligns widefilt output\n");
}
/*---------------------------------------------------------------------*/
short *set_rd_ptr(short flag)
{ /* done for legibility rather than speed */
  if(flag==1) return(strt1); else if (!flag) return(strt0);
  else err_rpt(" invalid flag");
  return(NULLPTR);
}
/*---------------------------------------------------*/
void main(int argc,char **argv) /* MAIN starts here. */
{
/* internal file variables */
  char    *inputname, *outputname, sntnc[60];
  FILE    *reading_from, *writing_to, *bq_file;
/* internal processing variables */
  long indx, dly_i, channel, delays[N_OF_CHANS], dlymax, dlymin, options;
  short rd_seg, offst, lastrd, nextrd, dummy = 0;
  short *oldptr, *newptr, *rdstrt, *tmprdstrt, *rdspace, *wrspace; /* data area pointers */

/* Parse Command Line */ options = 2;

  for (indx=1; indx < argc; indx++) {
    switch (*argv[indx]) {
      case '-':
	switch (*(argv[indx] + 1)) {
	  case 'i': inputname  = argv[++indx]; options--; break;
	  case 'o': outputname = argv[++indx]; options--; break;
	  case 'I': sendhelp(); break;
      }
    }
  }
  if (options) sendhelp();

/* Open all the Files and data space needed */
  if ( (reading_from = fopen (inputname, "rb")) == NULLPTR){
    strcpy(sntnc,inputname); err_rpt(" can't open this INPUT file\n");
  }

  if ( (writing_to = fopen (outputname, "wb")) == NULLPTR){
    strcpy(sntnc,outputname); err_rpt(" can't create this OUTPUT file\n");
  }

  if ( (bq_file = fopen ("biquad.tst", "rt")) == NULLPTR){
    printf(" falign: biquad.tst not available\n");exit(1);
  }

  if((rdspace = (short *) malloc(INSEG*N_OF_CHANS*sizeof(short))) == NULLPTR)
    err_rpt(" out of MEMORY for file read");

  if((wrspace = (short *) malloc(OUTSEG*N_OF_CHANS*sizeof(short)))== NULLPTR)
    err_rpt(" out of MEMORY for file write");

  indx = rd_biquad_file(bq_file, filter_coeffs, N_OF_CHANS); fclose(bq_file);

  if(indx != N_OF_CHANS){
    sprintf(sntnc," needed %2d, but only %2d BIQUAD filters",N_OF_CHANS,indx);
    err_rpt(sntnc);
  }

  filt_i = filter_coeffs;   /* use local pointer for legibility */
  dlymax = 0; dlymin = 255;

  printf("\nf_align: ");
  for(channel = 0; channel < N_OF_CHANS; channel++){
    dly_i = filt_i->delay; /* nominal delay; sort out later */
    if    (dly_i > dlymax) dlymax = dly_i;
    else if (dly_i < dlymin) dlymin = dly_i;
    delays[channel] = dly_i;
    filt_i++;   /* use local pointer for legibility */
  }
/* --------------- */
  for(channel = 0; channel < N_OF_CHANS; channel++) {
    delays[channel] -= dlymin; /* include check that this can handle length */
    if(delays[channel] > OUTSEG) {
      sprintf(sntnc," cannot handle shift of %3d",&delays[channel]); err_rpt(sntnc);
    }
	}
/* --------------- */
	strt0 = rdspace; strt1 = strt0 + (OUTSEG*N_OF_CHANS);
	if (( nextrd = fread (strt0, sizeof(short), (OUTSEG*N_OF_CHANS), reading_from)) < 0)
		err_rpt(" Error in READING data.");
	nextrd /= N_OF_CHANS; rd_seg = 1; lastrd = nextrd;  /* update markers */

	while(lastrd > 0){
		printf(".");fflush(stdout);
		rdstrt = set_rd_ptr(rd_seg);
		if(lastrd == OUTSEG) /* not yet at last block so perform another read */
			if (( nextrd = fread (rdstrt, sizeof(short), (N_OF_CHANS*OUTSEG), reading_from)) < 0)
				err_rpt(" Error in READING data.");

		nextrd/= N_OF_CHANS;
		rd_seg ^= 1; /* invert flag to say in which block to start shuffling */

		for(channel = 0; channel < N_OF_CHANS; channel++){ /* shift by each Y axis */
			dly_i = delays[channel];
			rdstrt = set_rd_ptr(rd_seg);
			for(indx = 0; indx < lastrd; indx++){ /* index by sample number */
				oldptr = wrspace + (N_OF_CHANS * indx) + channel; /* where data will go to */
				offst  = indx+dly_i; /* offset of where to get data from: perform checks */
				if(lastrd > offst) /* potential indexing out of current area: wrap into next */
					newptr = rdstrt +(N_OF_CHANS * offst) + channel;/* data came from */
				else {
					if((offst - nextrd - lastrd) > 0) newptr = &dummy; /* EoF close: stuff matrix */
					else {
						tmprdstrt = set_rd_ptr((rd_seg^1)); /* use other area temporarily */
						newptr = tmprdstrt + (N_OF_CHANS * (offst-OUTSEG)) + channel;
					}
				}
				*oldptr = *newptr; /* perform shift */
			}
		}
		if (fwrite (wrspace, sizeof(short), (lastrd*N_OF_CHANS), writing_to) == 0)
			err_rpt(" Error in WRITING data.");
		lastrd = nextrd; if(nextrd < OUTSEG) nextrd = 0; /* update and check last block */
	} /* end of while(lastrd>0) */

	fflush(writing_to);
	fclose(reading_from); fclose(writing_to);
  printf("\nf_align: NORMAL termination\n");
}
