/*----------------------------------------------------------------------

   nal_calc.c:  to calculate the insertion gains required of a hearing aid
		to meet the NAL from an audiogram containing frequencies
               .125, .25, .5, .75, 1, 1.5, 2, 3, 4, 6, & 8 kHz
                for the recruitment program channel frequencies
                (ref. Byrne & Dillon, Ear. hear. 7:257-265, 1986)

and for recruitment simulation, calcs the power N to which the envelope is
raised for each channel frequency. (assumes that loudness c'catches up' with
normal at level 'LEQUAL' dB, see below)
------------------------------------------------------------------------*/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NPTS_IN  11       /*  11 audiogram HL's */
#define NPTS_OUT 13       /*  13 filter channel frequencies */
#define LEQUAL   100.0    /*  point where normal / impaired ear have equal loudness */

float lin_int(float *yy,float *xx, float x,int N);
void locate(float *xx,int n,float x,int *j);

/*-------------------------------------------------------------------------------------*/
float lin_int(float *yy,float *xx, float x,int N)/* linear interpolation of parameters */
{
int j;

  locate(xx-1,N,x,&j);

  if(j < 1){
    return(yy[0]);
  }else if(j >= N){
    return(yy[N-1]);
  }else{
    return(yy[j-1]+(yy[j]-yy[j-1])*(x-xx[j-1])/(xx[j]-xx[j-1]));
  }
}

/*=============================================================================*/

void locate(float *xx,int n,float x,int *j) /*  Numerical Recipes  page 98 */
{
int ascnd,ju,jl,jm;

  jl = 0;
  ju = n + 1;
  ascnd=xx[n] > xx[1];
  while(ju-jl > 1) {
    jm = (ju + jl) >> 1;
    if((x > xx[jm]) == ascnd)
      jl = jm;
    else
      ju = jm;
  }
  *j = jl;

}
/*=============================================================================*/
void main(void)
{
int i;
float hl[NPTS_IN], X, gain, slope; /* slope is power to which envelope is raised */

float aud_consts[NPTS_IN] = { -18,-17.,-8.,-3.,1., 1.,-1.,-2.,-2.,-2.,-2.};
float aud_freqs [NPTS_IN] = {.125, .25, .5,.75,1.,1.5, 2., 3., 4., 6., 8.};

float rec_consts[NPTS_OUT]={-18.,-18.,-15.,-10.,-5.,-1.,1.,1.,-1.,-2.,-2.,-2.,-2};
float rec_freqs [NPTS_OUT]={.1,.19,.306,.452,.64,.879,1.184,1.572,2.067,2.698,3.503,4.529,5.837};
float rec_hl[NPTS_OUT];
char fileout[50];
int verbose = 1; /* printing to screen 0 = off  1 = on */
FILE *fd_out;

  printf("\nProgram to calculate insertion gains of hearing aids to NAL formulae\n");
  printf("Audiogram specified in dB(HL) at %2d frequencies\n",NPTS_IN);
  printf("Output file name for slopes and NAL gains?  ");
  scanf("%s",fileout);

/* open file */
  if((fd_out = fopen(fileout,"wt")) == NULL){
    printf("\n unable to open this file [%s]\n",fileout);
    exit(1);
  }

  for(i = 0; i < NPTS_IN; i++){
      printf("Enter the loss at %5.3fkHz: ",aud_freqs[i]);
      scanf("%f",&hl[i]);
  }

  X = 0.05 * (hl[2] + hl[4] + hl[6]);  /* part of NAL formula */


  if(verbose) {
    printf("\n\nNAL gains at Audiogram frequencies\n\n");
    printf("\nfreq  loss(dB HL)  gain(dB)\n");
  }
  for(i = 0; i < NPTS_IN; i++){
    gain =  X + 0.31 * hl[i] + aud_consts[i];
    if(verbose) printf("%5.3f,%7.0f,%10.1f\n",aud_freqs[i],hl[i], gain);
  }
  if(verbose) printf("\n");

/* use linear interpolation to get gains required at channel centre frequencies */
  for(i=0; i< NPTS_OUT; i++) rec_hl[i] = lin_int(hl,aud_freqs,rec_freqs[i],NPTS_IN);

/* N = 1/(1-hl/Lequal) */

  for(i=0; i < NPTS_OUT; i++){
    slope = 1./(1.-rec_hl[i]/LEQUAL);
    gain = X + 0.31 * rec_hl[i] + rec_consts[i];
    if(verbose) printf("%6.3f,  %4.0f, %5.2f,  %5.0f\n",rec_freqs[i],rec_hl[i],slope,gain);
    fprintf(fd_out,"%5.2f  %5.0f\n",slope,gain);
  }
  printf("\nData written to file %s\n",fileout);
}
