/*----------------------------------------------------------------------------

rlevel   - default output is level which would be obtained using Sennheiser
           headphones off the Masscomp with 10-dB attenuation
           i.e. +- 32767 sinusoid (rms = 32767/sqrt(2) = 23169.8) = 102 dB SPL 
	   or 60 dB has an rms 184.05, 80 dB 1840.5  [23170*10**(-22/20)]


     rlevel - Measures the mean rms level of a speech file, and measures
     the maximum and minimum sample values.

     rlevel [ -d max_dB -t threshold (-60 dB) -v verbose ] speechfile

  -d max_dB  max_dB is the level in dB SPL that a sinusoid with maximum
             levels of +-32767 would have. The default value is 102 dB


  -t dB    -  rejection threshold for mean rms calculations.  This is used
    to reject parts of the signal which contain silence from the
     calculation of the mean. This should be in dB and relative to 
    a maximum of 0dB. 0dB is considered to be the rms level of a
    sine wave which fits the 16 bit range exactly - ie
    +/- 32767.

  -v    - verbose - print information about level frame by frame

-----------------------------------------------------------------------------*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#define	SMALLANALYSIS	100	/* Look at file in 100 sample frames */
#define	SFRAMESIZE	10000	/* Hold 10000 samples at a time */
#define	RMS0dB16BIT	(32767./sqrt(2.0))
#define	SHORTS		2
#define	TRUE		1
#define	FALSE		0

int  dcoffset (short *start, int number);
float  meansquare (short* start, int number, int dcoffset1, int *min, int *max);
void error(char* s1, char* s2);

/*---------------------------------------------------------------------------*/
void main  (int argc, char **argv)
{

FILE *inputfp;
char *inputfile;
int  verbose = FALSE, finished;
int samplesread, i, framesabovethreshold;
int framesread;
int parameters,max,min;
float  rms0dB, threshold, rms, rmsthreshold, meansqr, summeansqrs=0.;
float speechrms, dB_rms = 102.0;
short speech[SFRAMESIZE];


  parameters = 1;  /* minimum of 1 parameter must be given */ 
  threshold  = -60.;

  /* Parse Command Line */

  for (i=1; i < argc; i++){
    switch (*argv[i]){
    case '-':
      switch (*(argv[i] + 1)){
      case 't': threshold = atof(argv[++i]); break;
      case 'v': verbose++; break;
      case 'd': dB_rms = atof(argv[++i]); break;
      }
    default:
    inputfile = argv[i]; 
    parameters--;
    }
  }

  if (parameters>0){
    fprintf(stderr," need at least an input file\n");
    error("use: rlevel [ -d max_dB (102) -t threshold (-60dB) -v verbose ] speechfile\n",NULL);
  }

  if ( (inputfp = fopen (inputfile, "rb")) == NULL)
    error (" Can't open Speech file: %s\n",inputfile);

  printf("%s %s\n",argv[0],inputfile);

  finished = 0;
  min = 32767;
  max = -32767;
  framesabovethreshold=0;
  framesread =0;

  rms0dB = RMS0dB16BIT;

  rmsthreshold = rms0dB * pow(10.,(threshold/20.));

  while (!finished){
    samplesread = fread (speech, SHORTS, SFRAMESIZE, inputfp);
    if (ferror (inputfp))
      error ("Error reading Speech file %s\n",inputfile);
    finished = feof (inputfp);
    i=0;
    while (i < (samplesread-SMALLANALYSIS) ){
      framesread++;

/* commented out - this is replaced by a function call which does NOT take out
the dc offset.
      meansqr = meansquare(&speech[i],SMALLANALYSIS,dcoffset(&speech[i],SMALLANALYSIS),&min,&max);
*/

      meansqr = meansquare(&speech[i],SMALLANALYSIS,0,&min,&max);
      rms = sqrt(meansqr);
      if (rms > rmsthreshold){
	summeansqrs += meansqr;
	if (verbose) printf("ABOVE %.1fdB Threshold : ",threshold);  
	framesabovethreshold++;
      } else 
	if (verbose) printf("Below %.1fdB Threshold : ",threshold);
      if (verbose)  printf("rms : %5.2f  Level : %5.2f\n",rms,20.* log10(rms/rms0dB));
      i += SMALLANALYSIS;
    }
  }

  printf("rlevel does NOT take out any dc\n");
  printf ("Number of 10ms frames above threshold = %d out of %d\n",framesabovethreshold,framesread);
  if (framesabovethreshold >0) {
    speechrms = sqrt(summeansqrs/(float) framesabovethreshold);
    printf ("RMS of speech above %.1f dBV threshold = \t%.1f\n",threshold,speechrms);
    printf("Power of signal (max %.1f) is \t\t\t%.1f dB\n", dB_rms,20. * log10 ( speechrms/rms0dB) + dB_rms);
  }
  printf("Max = %d  Min = %d .\n",max,min);
  printf("-----------------------------------------------------\n");
  fclose (inputfp);
}
/*---------------------------------------------------------------------------*/
float  meansquare (short* start, int number, int dcoffset1, int *min, int *max)
{
  int  i,sample;
  short  *ptr;
  float  sumsqrs;

  dcoffset1 = 0;
  sumsqrs =0.;
  ptr = start;
  for (i=0;i < number; i++){
    sample = *ptr;
    if (sample <*min) *min = sample;
    if (sample >*max) *max = sample;
    sample -= dcoffset1;
    sumsqrs += (sample * sample);
    ptr++;
  }
  return (sumsqrs/(float)number);
}
/*--------------------------------------------------------------------------*/

int  dcoffset (short *start, int number)
{
  int  i;
  short  *ptr;
  float  sum;

  sum  =0.;
  ptr = start;
  for (i=0;i < number; i++)
    sum += (float) *ptr++;
  return ( (int) (sum /(float)number) );
}
/*--------------------------------------------------------------------------*/

void error(char* s1,char* s2)
{
  printf("\n"); printf(s1,s2); printf("\n"); exit(1);
}

/*---------------------------------------------------------------------------*/
